/*  avr_prommer - prommer for AVR micro-controllers
 *  Copyright (C) 2000  Romain Lievin
 *   This program was originally written by { Martin Ohsmann, 10.4. 1997 }
 *   I translated it into C, ported to Linux and adapted for programming
 *   AT90S2313 also.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "lowlevel.h"
#include "avr_fnct.h"

/* Global vars */

int max_cnt = S1200_MAX_CNT;
int max_eep = S1200_MAX_EEP;

char flash_fname[MAXCHARS] = "rom.rom";
char eeprom_fname[MAXCHARS] = "eep.eep";
char auto_flash[MAXCHARS] = "off";
int adr_cnt_flash, adr_cnt_eeprom;
word addrs[MAX_CNT];
word datas[MAX_CNT];
byte eep[MAX_EEP];

/* File functions */

int get_flash_file(char *fname)
{
  FILE *f;
  word addr;
  word data;

  f=fopen(fname, "rt");
  if(f == NULL)
    {
      printf("Unable to open the FLASH file: %s\n", fname);
      return -1;
    }
  adr_cnt_flash = 0;
  while(!feof(f))
    {
      if(feof(f)) break;
      fscanf(f, "%06X:%04X\n", &addr, &data);
      addrs[adr_cnt_flash] = addr;
      datas[adr_cnt_flash] = data;
      adr_cnt_flash++;
    }
  printf("Flash file loaded: %i words\n", adr_cnt_flash);
  fclose(f);

  return 0;
}

int get_eep_file(char *fname)
{
  FILE *f;
  word addr;
  word data;
  int i;

  f=fopen(fname, "rt");
  if(f == NULL)
    {
      printf("Unable to open the EEPROM file: %s\n", fname);
      return -1;
    }
  for(i=0; i<max_eep; i++)
    eep[i]=0xFF;
  adr_cnt_eeprom = 0;
  while(!feof(f))
    {
      if(feof(f)) break;
      fscanf(f, "%04X:%02X\n", &addr, &data);
      eep[addr]=data;
      adr_cnt_eeprom++;
    }
  printf("Eeprom file loaded: %i words\n", adr_cnt_eeprom);
  //fprintf(stdout, "%04X: %02X\n", addr, eep[0]);
  fclose(f);

  return 0;
}

/* AVR functions */

void avr_enable(void)
{
  send_byte(0xAC);
  send_byte(0x53);
  send_byte(0x00);
  send_byte(0x00);
}

void avr_erase(void)
{
  printf("Erasing... ");
  issue_reset();
  avr_enable();

  send_byte(0xAC);
  send_byte(0x80);
  send_byte(0x00);
  send_byte(0x00);

  wtk(t100ms);
  issue_reset();
  printf("Done.\n");
}

byte avr_read_dev_code(int loc)
{
  send_byte(0x30);
  send_byte(0x00);
  send_byte(loc);
  return get_byte();
}

void avr_pgm_code(int loc, word w)
{
  send_byte(0x40);
  send_byte(hi(loc));
  send_byte(lo(loc));
  send_byte(lo(w));
  
  send_byte(0x48);
  send_byte(hi(loc));
  send_byte(lo(loc));
  send_byte(hi(w));

  wtk(t10ms);
}

word avr_read_code(int loc)
{
  byte b;
  word w;

  send_byte(0x20);
  send_byte(hi(loc));
  send_byte(lo(loc));
  w=get_byte();

  send_byte(0x28);
  send_byte(hi(loc));
  send_byte(lo(loc));
  b=get_byte();

  return  w + (b << 8);
}

byte avr_read_data(int loc)
{
  send_byte(0xA0);
  send_byte(0x00);
  send_byte(lo(loc));
  return get_byte();
}

void avr_pgm_data(int loc, int b)
{
  send_byte(0xC0);
  send_byte(0x00);
  send_byte(lo(loc));
  send_byte(b);
}

void avr_pgm_flash(void)
{
  int k;
 
  printf("Programming FLASH...\n");
  printf("Status: ");
  issue_reset();
  avr_enable();
  
  for(k=0; k<adr_cnt_flash; k++)
    {
      avr_pgm_code(addrs[k], datas[k]);
      printf(".");
    }
  printf("\n");
  printf("Done.\n");
}

void avr_pgm_eep(void)
{
  int k;

  printf("Programming EEPROM\n");
  printf("Status: ");
  issue_reset();
  avr_enable();

  //fprintf(stdout, "%04X %02X\n", 0, eep[0]);
  for(k=0; k<63; k++)
    {
      avr_pgm_data(k, eep[k]);
	  printf(".");
    }
  printf("\n");
  printf("Done.\n");
}

int exit_key(void)
{
  int c;
  char s[256];

  printf("\n");
  printf("Type 'c' to go on or 'x' to exit: ");
  fflush(stdin);
  fscanf(stdin, "%s", s);
  c = toupper(s[0]);
  if(c == 'X')
    return 1;
  else
    return 0;
}

void avr_show_flash(void)
{
  int k;

  issue_reset();
  avr_enable();
  printf("\n");

  for(k=0; k<512; k++)
    {
      if( ((k % 128) == 0) && (k>0) )
	{
	  if(exit_key()) return;
	}
      if((k % 8) == 0)
	{
	  printf("\n");
	  printf("%06X:", k);
	}
	printf(" %04X", avr_read_code(k));
      }
      printf("\n");
}

void avr_show_eep(void)
{
  int k;

  issue_reset();
  avr_enable();
  printf("\n");
  for(k=0; k<63; k++)
  {
    if( ((k % 128) == 0) && (k>0) )
    {
      if(exit_key()) return;
    }
    if((k % 16) == 0)
    {
      printf("\n");
      printf("%04X: ", k);
    }
    printf(" %02X", avr_read_data(k));
    avr_read_data(k);
  }
  printf("\n");
}

void avr_verify_flash(void)
{
  int k;
  word w;
  int errors;

  printf("\n");
  printf("Verifying FLASH...\n");
  issue_reset();
  avr_enable();
  errors=0;
  for(k=0; k<adr_cnt_flash; k++)
  {
    w=avr_read_code(addrs[k]);
    if(datas[k] != w)
    {
      errors++;
      if(errors < 5)
      printf(" @ %06X : should be %04X was %04X\n", addrs[k], datas[k],
	datas[k]);
    }
  }
  printf("-> %i errors\n", errors);
}

void show_device_code(void)
{
  int k;
  printf("\n");
  printf("DEVICE code: ");
  issue_reset();
  avr_enable();
  for(k=0; k<3; k++)
    printf("%02X ", avr_read_dev_code(k));
  printf("\n");
  //exit(0);
}
